#include "copyright.h"

//=================================================================
// img_dat.h - image data class templates.
//
// Description:
// image_t - Root image template class.
//           All images are derived from this class.
//
//
// Assumption: The pixel type T doesn't contain pointers.
//
//
// RCS information:
// ----------------
// $RCSfile: img_dat.h,v $
// $Revision: 1.2 $
// $Date: 1997/07/31 15:26:31 $
//
//
// copyright (c) 1997 The Hebrew University Of Jerusalem
// all rights reserved.
//=================================================================

#ifndef IMAGE_C_H
#define IMAGE_C_H

#include <string.h>
#include <assert.h>

//-----------------
// image_t defintion
//-----------------
template<class T> class image_t {
  
protected:

  unsigned width, height;
  T *data;

  int orgX, orgY;
  T *pix;                  // Relative to origin !

private:
  // Image manipulation
  void Realloc(const unsigned w, const unsigned h){
    unsigned size = w * h, oldsize;
    if ((oldsize = GetSize()) != size){
      if (oldsize != 0) 
        delete [] data;
      if (size != 0){
        data = new T[size];
        assert(data != NULL);
      }
      else data = (T *) 0;
    }
    width = w;
    height = h;
  }
  
  
public:  
  
  // Get attributes
  unsigned  GetWidth()  const  { return width; }
  unsigned  GetHeight() const  { return height;}
  unsigned  GetSize()   const  { return width * height; }
  const T * GetData()   const  { return data; }

  int GetOrgX()   const { return orgX; }
  int GetOrgY()   const { return orgY; }
  int GetXmin()   const { return  -orgX; }
  int GetYmin()   const { return  -orgY; }
  int GetXmax()   const { return  width  - orgX; }
  int GetYmax()   const { return  height - orgY; }
  


  // Set attributes.
  // Set new origin.  (origin must be inside the image)
  // note that Y coordinate positive direction is south ...
  void SetOrigin(const int x, const int y) { 
    orgX = x;
    orgY = y;
  
    if (data != (T *) 0) {
      assert ((orgX >= 0 && orgX < (int) width && 
	       orgY >= 0 && orgY < (int) height));
      pix = & data[y*width + x];
    }
    else pix = (T *) 0;
  }
  
  void MovOrigin(const int x, const int y) { 
    SetOrigin(orgX + x, orgY + y);
  }
  
  


  // indexing operator(s)
  
  //treat the image as a vector. index between 0 and GetSize()
  T &Pix(const unsigned i){
    assert(i < GetSize());
    return data[i];
  }

  const T &Pix(const unsigned i) const{
    assert(i < GetSize());
    return data[i];
  }

  // 2D relative to origin indexing operators  
  T &Pix(const int x, const int y){
    assert(x >= GetXmin() && x < GetXmax());
    assert(y >= GetYmin() && y < GetYmax());
    return pix[y*width + x];
  }

  const T &Pix(const int x, const int y) const{
    assert(x >= GetXmin() && x < GetXmax());
    assert(y >= GetYmin() && y < GetYmax());
    return pix[y*width + x];
  } 
  
  // Speical indexing operators
  
  T cPix (int x, int y) {
    if (x <  GetXmin()) x = GetXmin();
    else if (x >= GetXmax()) x = GetXmax() -1 ;
    
    if (y <  GetYmin()) y = GetYmin();
    else if (y >= GetYmax()) y = GetYmax() -1 ;
    
    return pix[y*width + x];
  }
  

  // Data initialization operator.
  image_t &operator=(const T val){
    for (unsigned i=0; i < GetSize(); i++)
      data[i] = val;
    return *this;
  }      
  
  // Assignment operator.
  image_t &operator=(const image_t &src){
    Realloc(src.width, src.height);
    width  = src.width;
    height = src.height;
    SetOrigin(src.orgX,src.orgY);

    
    memcpy((void *)data, (const void *) src.data, 
	   width*height*sizeof(T));
    
    return *this;
  }      
  
  
  // Constructor(s)
  image_t(){
    data = (T *) 0;		// null image
    width = 0;
    height = 0;
    pix  = (T *)0 ;
    orgX = 0;
    orgY = 0;
  }
  
  image_t(const unsigned w, const unsigned h, const int ox=0, const int oy=0){
    data = new T [w * h];
    assert(data != NULL);
    width = w;
    height = h;
    SetOrigin(ox, oy);

  }
  
  // copy constructor
  image_t(const image_t &img){
    width  = img.width;
    height = img.height;
    
    data = new T [width * height];
    assert(data !=NULL);
    memcpy((void *)data, (const void *) img.data, 
	   width*height*sizeof(T));

    SetOrigin(img.orgX, img.orgY);
  }
  
  // Distructor
  ~image_t(){
    char *tmp;
    tmp = (char *) data;
    if (GetSize() != 0) delete [] data ;
  }
};

#endif


