#include "copyright.h"

//=================================================================
// img_op.h - image operator class template.
//
// Description:
// img_t -  Incore root image template class.
//          All image manipulations are done on this class (and its
//          derivatives).
//
// Assumption: The pixel type T doesn't contain pointers. 
//
// RCS information:
// ----------------
// $RCSfile: img_op.h,v $
// $Revision: 1.3 $
// $Date: 1997/07/31 17:30:09 $
//
//=================================================================

#ifndef IMG_H
#define IMG_H

#include <string.h>
#include <assert.h>
#include <math.h>

#include "img_dat.h"

//----------
// fkernel
//----------

// Kernel images are used for convolution
// they degraded version of img_t

class fKernel : public image_t<float> {


 protected:
  using image_t<float>::width;
  using image_t<float>::height;
  using image_t<float>::data;
 
 public:
  using image_t<float>::GetWidth;
  using image_t<float>::GetHeight;
  using image_t<float>::GetSize;
  using image_t<float>::GetData;
  
  // disallow origin change
    void MovOrigin(const int, const int) {
      assert(("Cannot move origion of Kernel",0));
    }
  
  void SetOrigin(const int, const int){
    assert(("Cannot move origion of Kernel",0));
  }
  
  //(no casting of kernel)
    
    fKernel &operator=(const fKernel &ker){
      image_t<float>::operator=(ker);
      return *this;
    }
  
  // only convolution operator
  
  fKernel operator*(const fKernel &ker){
    fKernel tmp(width, height);
    
    for (register int x = GetXmin() ; x < GetXmax() ; x++){
      for (register int y = GetYmin() ; y < GetYmax() ; y++){
	for (register int i = ker.GetXmin() ; i < ker.GetXmax() ; i++){
	  for (register int j = ker.GetYmin() ; j < ker.GetYmax() ; j++){
	    //tmp.Pix(x,y) +=  cPix(x+i,y+j) * kernel.Pix(i,j)  ; 
	    tmp.Pix(x,y) +=  cPix(x-i,y-j) * ker.Pix(i,j)  ; 
	  }
	}
      }
    }
    return tmp;
  }

  // Kernel constructor
    fKernel(const unsigned w, const unsigned h, 
	    const float *init) : image_t<float>(w,h,w/2,h/2) {
	      memcpy((void *) data, (void *) init, w * h * sizeof(float));
	    }

    fKernel(const unsigned w, const unsigned h) : image_t<float>(w,h,w/2,h/2) {
	      memset((void *) data, '\0', w * h * sizeof(float));
	    }

};




//------------
// img_t
//------------

extern fKernel Di2Dx3, Di2Dy3, Di2Dx5, Di2Dy5, Di2Dx7, Di2Dy7;
extern fKernel Gx3, Gy3, Gx5, Gy5, Gx7, Gy7;

template<class T> class img_t : public image_t<T> {

public:

  using image_t<T>::width;
  using image_t<T>::height;
  using image_t<T>::orgX;
  using image_t<T>::orgY;
  using image_t<T>::pix;
  using image_t<T>::data;

  using image_t<T>::GetWidth;
  using image_t<T>::GetHeight;
  using image_t<T>::GetSize;
  using image_t<T>::GetData;
  using image_t<T>::GetXmin;
  using image_t<T>::GetYmin;
  using image_t<T>::GetXmax;
  using image_t<T>::GetYmax;
  using image_t<T>::Pix;
  using image_t<T>::cPix;
 
  
  // float subscript indexing - bilinear interpolation.
  T bPix(const float x, const float y) {
    int x1, x2;
    int y1, y2;
    float u, v;
    T a, b, c, d;
    T t1, t2, ret_val;
    
    x1 = (int) x;
    if (x < x1) x1--;
    x2 = x1 + 1;
    u = x - x1;
    
    y1 = (int) y;
    if (y < y1) y1--;
    y2 = y1 + 1;
    v = y - y1;
    
    a = cPix(x1, y1);
    b = cPix(x2, y1);
    c = cPix(x1, y2);
    d = cPix(x2, y2);
    
    t1 = a * (1 - u) + b * u;
    t2 = c * (1 - u) + d * u;
 
    ret_val = t1 * (1 - v) + t2 * v;

    return ret_val;
  }

  
  // -- reflexive Image scalar operators --
  
  img_t &operator=(const T val){
    image_t<T>::operator=(val);
    return *this;
  }      
  
  img_t &operator+=(const T val){
    for (register unsigned i=0; i<GetSize(); i++)
      data[i] += val;
    return *this;
  }
  
  img_t &operator-=(const T val){
    for (register unsigned i=0; i<GetSize(); i++)
      data[i] -= val;
    return *this;
  }
  
  img_t &operator*=(const T val){
    for (register unsigned i=0; i<GetSize(); i++)
      data[i] *= val;
    return *this;
  }
  
  img_t &operator/=(const T val){
    for (register unsigned i=0; i<GetSize(); i++)
      data[i] /= val;
    return *this;
  }
  
  // -- non reflexive Image scalar operators
  
  // need to define friend functions.
  // friend img_t &operator+<T>(const T &val, const img_t &img){
  // img.operator+(val);
  //}
  
  
  img_t operator+(const T val){
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize(); i++)
      tmp.data[i] = data[i] + val;
    return tmp;
  }
  
  img_t operator-(const T val){
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize(); i++)
      tmp.data[i] = data[i] - val;
    return tmp;
  }
  
  img_t operator*(const T val){
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize(); i++)
      tmp.data[i] = data[i] * val;
    return tmp;
  }
  
  img_t operator/(const T val){
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize(); i++)
      tmp.data[i] = data[i] / val;
    return tmp;
  }
  
  // reflexive Image <-> Image operators
  
  img_t &operator=(const img_t &img){
    image_t<T>::operator=(img);
    return *this;
  }      
  
  img_t &operator+=(const img_t &img){
    assert(width == img.width && height == img.height); 
    for (register unsigned i=0; i<(unsigned)GetSize(); i++)
      data[i] += img.data[i];
    return *this;
  }
  
  img_t &operator-=(const img_t &img){
    assert(width == img.width && height == img.height); 
    for (register unsigned i=0; i<(unsigned)GetSize(); i++)
      data[i] -= img.data[i];
    return *this;
  }
  
  img_t &operator*=(const img_t &img){
    assert(width == img.width && height == img.height); 
    for (register unsigned i=0; i<(unsigned)GetSize(); i++)
      data[i] *= img.data[i];
    return *this;
  }
  
  img_t &operator/=(const img_t &img){
    assert(width == img.width && height == img.height); 
    for (register unsigned i=0; i<(unsigned)GetSize(); i++)
      data[i] /= img.data[i];
    return *this;
  }
  
  // -- non reflexive Image <-> Image operators--
  
  img_t operator+(const img_t &img){
    assert(width == img.width && height == img.height); 
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize() ; i++)
      tmp.data[i] = data[i] + img.data[i];
    return tmp;
  }
  
  img_t operator-(const img_t &img){
    assert(width == img.width && height == img.height); 
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize() ; i++)
      tmp.data[i] = data[i] - img.data[i];
    return tmp;
  }
  
  img_t operator*(const img_t &img){
    assert(width == img.width && height == img.height); 
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize() ; i++)
      tmp.data[i] = data[i] * img.data[i];
    return tmp;
  }
  
  img_t operator/(const img_t &img){
    assert(width == img.width && height == img.height); 
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize() ; i++)
      tmp.data[i] = data[i] / img.data[i];
    return tmp;
  }
  
  // convolution - not optimized !
  
  // no const here ???
  img_t operator*(fKernel &kernel){
    img_t tmp(width, height, orgX, orgY);
    tmp = (T) 0.0F;
    
    for (register int x = GetXmin() ; x < GetXmax() ; x++){
      for (register int y = GetYmin() ; y < GetYmax() ; y++){
	for (register int i = kernel.GetXmin() ; i < kernel.GetXmax() ; i++){
	  for (register int j = kernel.GetYmin() ; j < kernel.GetYmax() ; j++){
	     //tmp.Pix(x,y) +=  cPix(x+i,y+j) * kernel.Pix(i,j)  ; 
	     tmp.Pix(x,y) +=  cPix(x-i,y-j) * kernel.Pix(i,j)  ; 
	  }
	}
      }
    }
    return tmp;
  }
  
  
  // ---unary operators---
  
  // Map functions.
  
  img_t map(T f(T x)){
    img_t tmp(width, height, orgX, orgY);
    for (register unsigned i = 0 ; i < GetSize() ; i++)
      tmp.data[i] = f(data[i]);
    return tmp;
  }
  
  img_t map(T f(int x, int y, T p)){
    img_t tmp(width, height, orgX, orgY);
    for (register int x = GetXmin() ; x < GetXmax() ; x++)
      for (register int y = GetYmin() ; y < GetYmax() ; y++)
	tmp.Pix(x,y) = f(x,y,Pix(x,y));
    return tmp;
  }
  
  // negate
  img_t operator-(){
    return (*this * (T)(-1.0F));
  }
  
  //need to add flip / transpose / rot 90 180 ***
  
  
  // Derivatives.
  img_t dx(const unsigned char filt_len = 5) { 
    switch(filt_len){
    case 3: return (*this * Di2Dx3);
    case 5: return (*this * Di2Dx5);
    case 7: return (*this * Di2Dx7);
    default: 
      assert(("Dx kernel size not supported",0));
      return *this; //***
    }
  }
  
  img_t dy(const unsigned char filt_len = 5) { 
    switch(filt_len){
    case 3: return (*this * Di2Dy3);
    case 5: return (*this * Di2Dy5);
    case 7: return (*this * Di2Dy7);
    default: 
      assert(("Dx kernel size not supported",0));
      return *this; //***
    }
  }
  
  
  // Blure
  img_t gx(const unsigned char filt_len = 5) { 
    switch(filt_len){
    case 3: return (*this * Gx3);
    case 5: return (*this * Gx5);
    case 7: return (*this * Gx7);
    default: 
      assert(("Gx kernel size not supported",0));
      return *this; //***
    }
  }
  
  img_t gy(const unsigned char filt_len = 5) { 
    switch(filt_len){
    case 3: return (*this * Gy3);
    case 5: return (*this * Gy5);
    case 7: return (*this * Gy7);
    default: 
      assert(("Gx kernel size not supported",0));
      return *this; //***
    }
  }
  
  //*****************************************
  //*****************************************
  // contructor/distructor
  
  img_t() : image_t<T>(){}
  
  img_t(const unsigned w, const unsigned h, 
	const int ox=0, const int oy=0) : image_t<T>(w, h,ox,oy){}
  
  img_t(const img_t &img) : image_t<T>(img){}
  
  
  // image_t distructor is called to free data.
  
};


#endif
  
  
  
  
  
  
  
