
// Color compressed movie class

#include <fcntl.h>

//#define WINDOWS

#ifdef WINDOWS
#	include <winsock.h>
#	include	<io.h>
#	include <sys/stat.h>
#	define	open			_open
#	define	close			_close
#	define	read			_read
#	define	write			_write
#	define	tell			_tell
#	define	lseek			_lseek
#	define  MODE			(_S_IREAD | _S_IWRITE)
#else
#	include <sys/types.h>
#	include <unistd.h>
#       include <netinet/in.h>
#	define	_O_SEQUENTIAL		0
#	define  _O_BINARY		0
#	define	_O_RDONLY		O_RDONLY
#	define	_O_WRONLY		O_WRONLY
#	define	_O_TRUNC		O_TRUNC
#	define	_O_CREAT		O_CREAT
#	define	MODE			0644
#	define tell(x)			lseek((x),SEEK_CUR,0L)
#endif

#include "image.h"
#include "movie.h"
#include "error_msg.h"


//---------------------
// zlib functions
//---------------------
void checkver()
{
	if (zlibVersion()[0] != ZLIB_VERSION[0]) {
		error_msg("incompatible zlib version\n",0);
		clip_exception(100);
	} 
	else if (strcmp(zlibVersion(), ZLIB_VERSION) != 0) {
		error_msg("warning: different zlib version\n",0);
	}
}

void check_err(int err, char *msg) 
{ 
	if (err != Z_OK) { 
		error_msg(msg, err); 
		clip_exception(err); 
	} 
}



//------------------------
// Movie class functions
//------------------------

//----------------------
// create fmovie file
//----------------------
void Movie::Create(char *fname)
{
	if ((fd = ::open(fname, _O_CREAT | _O_WRONLY | _O_TRUNC | _O_SEQUENTIAL | _O_BINARY, MODE))== -1) {
		error_msg("Can't create %s\n", (int)fname);
		clip_exception(OPEN);
	}
	
	write(fd,&header,sizeof(header));		// save space.

	dir_size = INITIAL_FRAME_NO;
	dir		   = new unsigned long[dir_size];
	assert(dir != NULL);


}

//---------------------
// add frames
//---------------------
float Movie::Add_frame(bRgbImg &img)
{
    int err;
	Byte *data;
	uLong datalen, outlen, size;


	if ((header.width == 0) || (header.height == 0)) {
		header.width  = img.GetWidth();
		header.height = img.GetHeight();
	}

	assert(header.width  == img.GetWidth());
	assert(header.height == img.GetHeight());

	data = (unsigned char *) img.GetData();
	size = img.GetSize();
	datalen = size * sizeof(bRgbPix);
	outlen = MAX_FRAME_SIZE;
	
	for (unsigned i=0; i<size; i++){
		chan[i]        = data[3*i];
		chan[i+size]   = data[3*i+1];
		chan[i+2*size] = data[3*i+2];
	}

	err = compress((Byte *) buf, &outlen, (Byte *) chan, datalen);
   	check_err(err, "compress");
	
	if (cur_frame >= dir_size) {
		unsigned long *tmp = dir;
		dir	= new unsigned long[dir_size*2];
		assert(dir != NULL);
		memcpy(dir,tmp,dir_size*sizeof(long));
		delete [] tmp;
		dir_size *=2;
	}
	dir[cur_frame++] = tell(fd);
	write(fd,buf,outlen);
	
	return (float) outlen * 100.0F / (float) datalen;
}

//-----------------
// open
//-----------------

void Movie::Open(char *fname)
{
	if ((fd = open(fname, _O_RDONLY | _O_SEQUENTIAL | _O_BINARY))== -1) {
		error_msg("Can't open %s\n", (int)fname);
		clip_exception(OPEN);
	}
	
	read(fd,&header,sizeof(header));
 
    if (strncmp(header.magic,"Clip_Mov",8)){
            error_msg("Unknown format %s\n", (int)fname);
            clip_exception(UNKNOWN_FORMAT);
    }

	header.width	   = ntohs(header.width);
	header.height	   = ntohs(header.height);
	header.dir_offset  = ntohl(header.dir_offset);
	header.frame_count = ntohl(header.frame_count);
	dir_size = header.frame_count+1;
	dir		   = new unsigned long[dir_size];
	assert(dir != NULL);

	lseek(fd, header.dir_offset, SEEK_SET);
	read(fd,dir,sizeof(long) * (header.frame_count+1));

	for (unsigned i=0; i<=header.frame_count; i++)
		dir[i] = ntohl(dir[i]);

	lseek(fd, dir[0], SEEK_SET);
	cur_frame = 0;
}


//------------------
// get frame
//------------------
void	Movie::Get_frame(bRgbImg &img, long fno)
{
	int err;
	Byte *outdata;
	uLong datalen, outlen, size;

	if (img.GetSize() == 0) img = bRgbImg(header.width, header.height);

	assert(header.width  == img.GetWidth());
	assert(header.height == img.GetHeight());
	assert((fno >= 0) && (fno < (long) header.frame_count));

	datalen = dir[fno+1] - dir[fno];

	outdata = (unsigned char *) img.GetData();
	size = img.GetSize();
	outlen  = size * sizeof(bRgbPix);
	
	if (cur_frame != (unsigned) fno) lseek(fd, dir[fno], SEEK_SET);
	read(fd,buf,datalen);
	cur_frame = fno+1;

	err = uncompress((Byte *) chan, &outlen, (Byte *) buf, datalen);
   	check_err(err, "un-compress");

	for (unsigned i=0; i<size; i++){
		outdata[3*i]   = chan[i];
		outdata[3*i+1] = chan[i+size] ;
		outdata[3*i+2] = chan[i+2*size];
	}

}

//-----------------
// Constructor(s)
//-----------------
Movie::Movie() {
	checkver();			// check zlib version integrety
	memset(header.magic,'\0',16);
	strcpy(header.magic,"Clip_Mov V1.0\n");
	header.width  = 0;
	header.height = 0;
	header.dir_offset  = 0;
	header.frame_count = 0;
	cur_frame  = 0;
	dir_size =   0;
	buf		   = new unsigned char[MAX_FRAME_SIZE];
	chan	   = new unsigned char[MAX_FRAME_SIZE];
	fd = -1;
}

//------------------
// destructor
//------------------
Movie::~Movie(){
	if (fd != -1) {
		
		// last location.
		dir[cur_frame] = header.dir_offset = tell(fd);
		header.frame_count = cur_frame;

		// convert to BIG ENDIAN before writing.
		for (unsigned i=0; i<=header.frame_count; i++)
			dir[i] = htonl(dir[i]);
		
		write(fd,dir,sizeof(long) * (header.frame_count+1));

		header.width	   = htons(header.width);
		header.height	   = htons(header.height);
		header.dir_offset  = htonl(header.dir_offset);
		header.frame_count = htonl(header.frame_count);
	
		lseek(fd, 0L, SEEK_SET);
		write(fd,&header,sizeof(header));
		close(fd);
	}
	delete [] dir;
	delete [] buf;
	delete [] chan;
}
